#!/usr/bin/env python
#
#   genann/genetics.py
#   genann
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Genetics"""

from itertools import izip
import random

__author__ = 'Ross Light'
__date__ = 'June 28, 2008'
__all__ = ['Gene',
           'NumericGene',
           'IntegerGene',
           'BooleanGene',
           'FloatGene',
           'Genome',
           'roulette',
           'population_step',]

class Gene(object):
    """
    A single gene inside a `Genome`.
    
    This is the base class for other genes, and should not be used directly.
    
    All subclasses must:
    * Be immutable
    * Be hashable
    """
    def get_value(self):
        """
        Accesses the value of the gene.
        
        :Returns: The gene's value.
        :ReturnType: object
        """
        raise NotImplementedError()
    
    def point_mutate(self):
        """
        Creates a mutated gene.
        
        :Returns: The modified gene
        :ReturnType: `Gene`
        """
        raise NotImplementedError()

class NumericGene(Gene):
    __slots__ = ['_value']
    _factory = None
    
    def __init__(self, value):
        self._value = self._factory(value)
    
    def get_value(self):
        return self._value
    
    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self._value)
    
    def __str__(self):
        return str(self._value)
    
    def __nonzero__(self):
        return bool(self._value)
    
    def __hash__(self):
        return hash(self._value)
    
    def __eq__(self, other):
        if isinstance(other, NumericGene):
            return self._value == other._value
        elif isinstance(other, (bool, float, int, long)):
            return self._value == other
        else:
            return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, NumericGene):
            return self._value != other._value
        elif isinstance(other, (bool, float, int, long)):
            return self._value != other
        else:
            return NotImplemented
    
    def __gt__(self, other):
        if isinstance(other, NumericGene):
            return self._value > other._value
        elif isinstance(other, (bool, float, int, long)):
            return self._value > other
        else:
            return NotImplemented
    
    def __lt__(self, other):
        if isinstance(other, NumericGene):
            return self._value < other._value
        elif isinstance(other, (bool, float, int, long)):
            return self._value < other
        else:
            return NotImplemented

class IntegerGene(NumericGene):
    __slots__ = []
    _factory = int

class BooleanGene(NumericGene):
    __slots__ = []
    _factory = bool
    
    def point_mutate(self):
        return BooleanGene(not self._value)

class FloatGene(NumericGene):
    __slots__ = []
    _factory = float

class Genome(object):
    __slots__ = ['_genes']
    
    def __init__(self, genes):
        self._genes = tuple(genes)
    
    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self._genes)
    
    def __str__(self):
        return "<%s>" % (', '.join(str(gene) for gene in self))
    
    def __iter__(self):
        return iter(self._genes)
    
    def __len__(self):
        return len(self._genes)
    
    def __getitem__(self, item):
        return self._genes[item]
    
    def __contains__(self, item):
        return item in self._genes
    
    def __hash__(self):
        return hash(self._genes)
    
    def __eq__(self, other):
        if isinstance(other, Genome):
            return self._genes == other._genes
        elif isinstance(other, (list, tuple)):
            return self._genes == other
        else:
            return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, Genome):
            return self._genes != other._genes
        elif isinstance(other, (list, tuple)):
            return self._genes != other
        else:
            return NotImplemented
    
    def __lt__(self, other):
        raise TypeError("Genomes can only be checked for equality")
    
    def __le__(self, other):
        raise TypeError("Genomes can only be checked for equality")
    
    def __gt__(self, other):
        raise TypeError("Genomes can only be checked for equality")
    
    def __ge__(self, other):
        raise TypeError("Genomes can only be checked for equality")
    
    def __add__(self, other):
        if isinstance(other, Genome):
            return self.one_point_crossover(self, other)
        else:
            return NotImplemented
    
    ## CROSSOVER OPERATORS ##
    
    @classmethod
    def one_point_crossover(cls, g1, g2):
        # Check parameters
        if len(g1) != len(g2):
            raise ValueError("Genomes' lengths must match")
        if g1 == g2 or len(g1) < 2:
            return g1, g2
        # Choose crossover point
        crossover_point = random.randrange(len(g1))
        # Create children
        child1, child2 = [], []
        for i, (gene1, gene2) in enumerate(izip(g1, g2)):
            if i < crossover_point:
                # Verbatim copy
                child1.append(gene1)
                child2.append(gene2)
            else:
                # Cross-over point reached, swap genes
                child1.append(gene2)
                child2.append(gene1)
        return cls(child1), cls(child2)
    
    @classmethod
    def two_point_crossover(cls, g1, g2):
        # Check parameters
        if len(g1) != len(g2):
            raise ValueError("Genomes' lengths must match")
        if g1 == g2 or len(g1) < 2:
            return g1, g2
        # Choose crossover points
        index1 = random.randrange(len(g1))
        index2 = index1
        while index2 != index1:
            index2 = random.randrange(len(g1))
        index1, index2 = min(index1, index2), max(index1, index2)
        # Create children
        child1, child2 = [], []
        for i, (gene1, gene2) in enumerate(izip(g1, g2)):
            if i < index1 or i > index2:
                # Verbatim copy
                child1.append(gene1)
                child2.append(gene2)
            else:
                # Inside cross-over region, swap genes
                child1.append(gene2)
                child2.append(gene1)
        return cls(child1), cls(child2)
    
    ## MUTATION OPERATORS ##
    
    def point_mutate(self, rate):
        new_genes = []
        for gene in self:
            if random.random() < rate:
                gene = gene.point_mutate()
            new_genes.append(gene)
        return Genome(new_genes)
    
    def exchange_mutate(self):
        if len(self) < 2:
            return
        pos1 = pos2 = random.randrange(len(self))
        while pos2 == pos1:
            pos2 = random.randrange(len(self))
        new_genes = list(self)
        new_genes[pos1], new_genes[pos2] = new_genes[pos2], new_genes[pos1]
        return Genome(new_genes)

def roulette(input_list, weights):
    """
    A roulette wheel selection, based on weights.
    
    Items with larger weights are more likely to be picked.
    
    :Parameters:
        input_list : list
            A list of objects to select from
        weights : list of float
            Weights for probability of the input set
    :Returns: The selected object
    :ReturnType: 
    """
    # Check inputs
    if not input_list:
        raise ValueError("Cannot select from an empty set")
    assert len(input_list) == len(weights), "Incorrect weight count"
    # Perform selection
    target = random.random() * sum(weights)
    current_total = 0.0
    for obj, weight in izip(input_list, weights):
        current_total += weight
        if current_total > target:
            return obj
    else:
        raise AssertionError("Weights did not add up correctly")

def population_step(population, fitness_func,
                    selection_func=None,
                    crossover_func=None,
                    mutate_func=None,
                    elites=0,
                    elite_copies=1,):
    """
    Create a new population.
    
    :Parameters:
        population : list of `Genome`
            A population of genomes
        fitness_func : callable
            A function that takes a genome and returns its fitness.
    :Keywords:
        selection_func : callable
            A function that selects from a population.  The function should
            take two arguments: the population, and the fitness weights.  By
            default, this uses the `roulette` function.
        crossover_func : callable
            A function that can cross over two genomes.  If not given, no
            cross-over will take place.
        mutate_func : callable
            A function that can mutate a genome.  If not given, no mutation
            will take place.
        elites : int
            The number of elites to take from each generation.  By default,
            elitism is disabled.
        elite_copies : int
            The number of copies of the elites to be made.  By default, no
            additional copies are made.
    :Returns: A new population
    :ReturnType: list of `Genome`
    """
    # Check parameters
    if fitness_func is None:
        raise ValueError("Fitness function must be specified")
    if selection_func is None:
        selection_func = roulette
    # Start population
    new_population = []
    fitness = [fitness_func(genome) for genome in population]
    # Perform elitism
    if elites > 0:
        elite_list = sorted(population,
                            key=(lambda g: fitness[population.index(g)]),
                            reverse=True)
        elite_members = elite_list[:elites]
        new_population += elite_members * elite_copies
    # Create mates
    while len(new_population) < len(population):
        # Select parents
        parent1 = selection_func(population, fitness)
        parent2 = selection_func(population, fitness)
        # Crossover
        if crossover_func is None:
            children = [parent1, parent2]
        else:
            children = list(crossover_func(parent1, parent2))
        # Mutation
        if mutate_func is not None:
            children = [mutate_func(child) for child in children]
        # Add to population
        new_population += children
    return new_population
