#!/usr/bin/env python
#
#   genann/neuralnet.py
#   genann
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Feedforward neural network"""

from __future__ import absolute_import, division
from itertools import izip
import math
import random

from . import genetics, util

__author__ = 'Ross Light'
__date__ = 'June 28, 2008'
__all__ = ['Neuron',
           'NeuralNet',
           'NeuronGene',
           'NeuralGenome',]

class Neuron(object):
    __slots__ = ['_weights', '_bias', '_activator']
    
    @staticmethod
    def step_activator(isum):
        """
        Simple boolean activator.
        
        The function is defined mathematically as follows::
        
            f(x) = {1, x >= 1
                   {0, x < 1
        
        :Parameters:
            isum : float
                The input sum
        :Returns: The activation
        :ReturnType: float
        """
        return 1.0 if isum >= 1.0 else 0.0
    
    @staticmethod
    def make_logistic_activator(curvature=1.0):
        """
        Creates a logistic activator.
        
        The function is defined mathematically as follows::
        
            f(x) = 1 / (1 + e ^ (-x / n))
        
        :Parameters:
            curvature : float
                The curvature of the function.  Corresponds to "n" in the
                function.
        :Returns: The activator function
        :ReturnType: callable
        """
        def logistic_activator(isum):
            return 1 / (1 + math.e ** (-isum / curvature))
        return logistic_activator
    
    def __init__(self, weights, bias=0.0, activator=None):
        self._weights = tuple(weights)
        self._bias = float(bias)
        if activator is None:
            activator = self.step_activator
        self._activator = activator
    
    def __repr__(self):
        return "Neuron(%r, %r)" % (list(self._weights), self._bias)
    
    def __str__(self):
        return "{%s} (bias=%.3f)" % (', '.join(str(w) for w in self._weights),
                                     self._bias)
    
    def activate(self, inputs):
        if len(inputs) != len(self._weights):
            raise ValueError("Neuron requires %i inputs" % len(self._weights))
        activation_sum = 0.0
        for value, weight in izip(inputs, self._weights):
            activation_sum += value * weight
        activation_sum -= self._bias
        return self._activator(activation_sum)
    
    @property
    def weights(self):
        return self._weights
    
    @property
    def bias(self):
        return self._bias
    
    def __hash__(self):
        return hash(self._weights) ^ hash(self._bias)
    
    def __eq__(self, other):
        if isinstance(other, Neuron):
            return (self._weights == other._weights and
                    self._bias == other._bias)
        else:
            return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, Neuron):
            return (self._weights != other._weights or
                    self._bias != other._bias)
        else:
            return NotImplemented
    
    def __lt__(self, other):
        raise TypeError("Neurons can only be checked for equality")
    
    def __le__(self, other):
        raise TypeError("Neurons can only be checked for equality")
    
    def __gt__(self, other):
        raise TypeError("Neurons can only be checked for equality")
    
    def __ge__(self, other):
        raise TypeError("Neurons can only be checked for equality")

class NeuralNet(object):
    __slots__ = ['_layers']
    
    def __init__(self, layers):
        if not layers:
            raise ValueError("Must provide some layers")
        if isinstance(layers[0], Neuron):
            # Flat layer given
            self._layers = (tuple(layers),)
        else:
            # Multiple layers given
            if not all(layers):
                raise ValueError("Empty layer given")
            self._layers = tuple(tuple(layer) for layer in layers)
    
    def __repr__(self):
        return "NeuralNet(%r)" % ([list(layer) for layer in self._layers])
    
    def __str__(self):
        layer_strings = []
        for layer in self._layers:
            neuron_strings = [str(neuron) for neuron in layer]
            layer_str = '[%s]' % ('\n  '.join(neuron_strings))
            layer_strings.append(layer_str)
        return "[%s]" % ('\n '.join(layer_strings))
    
    def evaluate(self, inputs):
        for layer in self._layers:
            inputs = [neuron.activate(inputs) for neuron in layer]
        return inputs
    
    @property
    def layers(self):
        return self._layers

class NeuronGene(genetics.Gene):
    """A gene for a neuron."""
    __slots__ = ['_weights', '_bias']
    
    @classmethod
    def from_neuron(cls, neuron):
        return cls(neuron.weights, neuron.bias)
    
    def __init__(self, weights, bias=0.0):
        self._weights = tuple(weights)
        self._bias = float(bias)
    
    def get_value(self):
        return self._weights + (self._bias,)
    
    def point_mutate(self, rate, max_perturbation):
        def _mutate_weight(weight):
            return weight + util.random_delta() * max_perturbation
        limited_mutate = util.limit_function(_mutate_weight, rate)
        new_weights = [limited_mutate(weight) for weight in self._weights]
        bias = limited_mutate(self._bias)
        return NeuronGene(new_weights, bias)
    
    def to_neuron(self, activator=None):
        return Neuron(self._weights, self._bias, activator)
    
    def __repr__(self):
        return "NeuronGene(%r, %r)" % (list(self._weights), self._bias)
    
    def __str__(self):
        return "{%s} (bias=%.3f)" % (', '.join(str(w) for w in self._weights),
                                     self._bias)
    
    def __nonzero__(self):
        return bool(self._value)
    
    def __hash__(self):
        return hash(self._value)
    
    def __eq__(self, other):
        if isinstance(other, NeuronGene):
            return (self._weights == other._weights and
                    self._bias == other._bias)
        else:
            return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, NeuronGene):
            return (self._weights != other._weights or
                    self._bias != other._bias)
        else:
            return NotImplemented
    
    def __lt__(self, other):
        raise TypeError("Neuron gene can only be checked for equality")
    
    def __le__(self, other):
        raise TypeError("Neuron genes can only be checked for equality")
    
    def __gt__(self, other):
        raise TypeError("Neuron genes can only be checked for equality")
    
    def __ge__(self, other):
        raise TypeError("Neuron genes can only be checked for equality")

class NeuralGenome(genetics.Genome):
    __slots__ = []
    
    @classmethod
    def from_net(cls, net):
        genes = []
        for layer in net.layers:
            for neuron in layer:
                genes.append(NeuronGene.from_neuron(neuron))
        return cls(genes)
    
    def to_net(self, counts, activators=None):
        if sum(counts) != len(self):
            raise ValueError("Mismatched neuron layer counts")
        layers = []
        current_gene = 0
        for i, neuron_count in enumerate(counts):
            # Find activator
            if activators is None:
                activator = None
            else:
                activator = activators[i]
            # Fetch genes
            genes = self[current_gene:current_gene + neuron_count]
            current_gene += neuron_count
            # Create layer
            new_layer = [gene.to_neuron(activator) for gene in genes]
            layers.append(new_layer)
        return NeuralNet(layers)
