#!/usr/bin/env python
#
#   genann/util.py
#   genann
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Miscellaneous utility functions"""

from __future__ import absolute_import
import functools
import random

__author__ = 'Ross Light'
__date__ = 'June 29, 2008'
__all__ = ['limit_by',
           'limit_function',
           'random_delta',]

def limit_by(rate):
    """
    Decorator for `limit_function`.
    
    :Parameters:
        rate : float
            A number between zero and one that determines the calling rate
    :Returns: A wrapper that must be called with a function
    :ReturnType: callable
    """
    def wrapper(func):
        _limited_func = limit_function(func, rate)
        return functools.update_wrapper(_limited_func, func)
    return wrapper

def limit_function(func, rate):
    """
    Create a function that limits a function to a rate.
    
    If the function will not be called, then the value passed to the function
    will be passed through.
    
    :Parameters:
        func : callable
            The function to limit
        rate : float
            A number between zero and one that determines the calling rate
    :Returns: The limited function
    :ReturnType: callable
    """
    def _limited_func(*args):
        if random.random() < rate:
            return func(*args)
        else:
            if len(args) == 1:
                return args[0]
            else:
                return args
    return _limited_func

def random_delta():
    """
    Generates a random number between -1 and 1, exclusive.
    
    :Returns: A random delta
    :ReturnType: float
    """
    n = random.random()
    if random.randint(0, 1):
        return n
    else:
        return -n
