#!/usr/bin/env python
#
#   zfilenet/util.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Various utility classes and functions

:Variables:
    LOG_USAGE : int
        Constant for usage information
"""

from __future__ import absolute_import, division
from datetime import datetime, timedelta, tzinfo
import logging
import re
from uuid import UUID, uuid4

__author__ = 'Ross Light'
__date__ = 'June 13, 2008'
__all__ = ['Path',
           'ProgressDeferred',
           'BelowLevelFilter',
           'LOG_USAGE',
           'setup_log',
           'setup_log_options',
           'add_log_options',
           'to_uuid',
           'data2hex',
           'hex2data',
           'parse_qname',
           'format_file_size',
           'format_datetime',]

### PATHS ###

class Path(object):
    """
    A POSIX path conforming object.
    
    Paths are split up into components and are automatically normalized.
    
    :CVariables:
        sep : str
            Separator for string-based paths
        curdir : str
            Current directory indicator for string-based paths
        pardir : str
            Parent directory indicator for string-based paths
    :IVariables:
        components : tuple
            The components of the path
        absolute : bool
            Whether the path is absolute
        directory : bool
            Whether the path refers to a directory
    """
    __slots__ = ['components', 'absolute', 'directory']
    
    sep = '/'
    curdir = '.'
    pardir = '..'
    
    def __init__(self,
                 path=[],
                 absolute=False,
                 directory=False):
        """
        Create a path.
        
        :Parameters:
            path
                The method of creation is dependent on what type of object this
                is.  If it is a:
                * `Path` object, then a copy is created.
                * string, then the path is parsed from the string.
                * sequence, then the path's components are constructed from the
                  members of the sequence and the keywords specify what kind of
                  path it is.
        :Keywords:
            absolute : bool
                Whether the path is absolute (i.e. has a leading slash)
            directory : bool
                Whether the path is a directory (i.e. has a trailing slash)
        """
        if isinstance(path, Path):          # Copy initialization
            components = path.components
            self.absolute = path.absolute
            self.directory = path.directory
        elif isinstance(path, basestring):  # Initialization from string
            # Determine absoluteness
            if path.startswith(self.sep):
                self.absolute = True
                path = path[len(self.sep):]
            else:
                self.absolute = False
            # Determine directory
            if path.endswith(self.sep):
                self.directory = True
                path = path[:-len(self.sep)]
            else:
                self.directory = False
            # Split components
            components = path.split(self.sep)
        else:                               # Initialization from sequence
            components = path
            self.absolute = bool(absolute)
            self.directory = bool(directory)
        self.components = tuple(components)
        self._normpath()
    
    def _normpath(self):
        """
        Normalize the path.
        
        This involves removing empty components and resolving immediately
        solvable parent references (e.g. "foo/../bar" turns into "bar", but
        "../foo/bar" remains the same).
        """
        # Do immediate path cleaning
        result = []
        for component in self.components:
            if self.sep in component:
                raise ValueError("Separators are not allowed inside "
                                 "components")
            elif component and component != self.curdir:
                result.append(component)
        # Remove any fixable parent directory issues
        start_index = 0
        while True:
            try:
                parent_index = result.index(self.pardir, start_index)
            except ValueError:
                break
            else:
                previous_index = parent_index - 1
                if previous_index < start_index:
                    start_index += 1
                else:
                    del result[parent_index]
                    del result[previous_index]
        # Remove any leading parent directories if we are absolute
        if self.absolute:
            while result and result[0] == self.pardir:
                del result[0]
            if self.directory and not result:
                self.directory = False
        # Change components
        self.components = tuple(result)
    
    # Public API
    
    def relative_path(self, other):
        """
        Evaluate a path as relative to the caller.
        
        >>> path1 = Path('/home/python/')
        >>> path2 = Path('spam/eggs')
        >>> path1.relative_path(path2)
        Path(['home', 'python', 'spam', 'eggs'], absolute=True)
        >>> path2.relative_path('hello')
        Path(['spam', 'hello'])
        
        :Parameters:
            other : `Path`
                The path to resolve, relative to self
        :Returns: The resolved path
        :ReturnType: `Path`
        """
        other_path = Path(other)
        if other_path.absolute:
            return other_path
        elif self.directory:
            return self + other_path
        else:
            return self[:-1] + other_path
    
    def convert(self, **kw):
        """
        Convert path to a different type.
        
        This doesn't affect the components, only the type of path.
        
        :Keywords:
            absolute : bool
                Whether the new path should be absolute
            directory : bool
                Whether the new path should be a directory
        :Returns: The converted path
        :ReturnType: `Path`
        """
        parameters = {'absolute': self.absolute,
                      'directory': self.directory,}
        parameters.update(kw)
        return Path(self.components, **parameters)
    
    def sanitize(self, chars):
        """
        Sanitize the path by removing characters.
        
        :Parameters:
            chars : str
                The set of characters to remove
        :Returns: The sanitized path
        :ReturnType: `Path`
        """
        # Assemble regular expression
        pattern = '|'.join(re.escape(char) for char in chars)
        pattern = re.compile(pattern)
        # Sanitize components
        new_components = []
        for component in self.components:
            new_components.append(pattern.sub('', component))
        # Create new path
        return Path(new_components,
                    absolute=self.absolute,
                    directory=self.directory)
    
    # String representation
    
    def __repr__(self):
        result = "Path(%r" % list(self.components)
        if self.absolute:
            result += ", absolute=%r" % (self.absolute)
        if self.directory:
            result += ", directory=%r" % (self.directory)
        result += ")"
        return result
    
    def __str__(self):
        result = self.sep.join(self.components)
        if self.absolute:
            result = self.sep + result
        if self.directory:
            result += self.sep
        return result
    
    # Operators
    
    def __add__(self, other):
        if isinstance(other, Path):
            return Path(self.components + other.components,
                        absolute=self.absolute,
                        directory=other.directory)
        elif isinstance(other, basestring):
            return self + Path(other)
        elif isinstance(other, (tuple, list)):
            return Path(self.components + tuple(other),
                        absolute=self.absolute)
        else:
            return NotImplemented
    
    def __radd__(self, other):
        if isinstance(other, basestring):
            return Path(other) + self
        elif isinstance(other, (tuple, list)):
            return Path(tuple(other) + self.components,
                        directory=self.directory)
        else:
            return NotImplemented
    
    def __len__(self):
        return len(self.components)
    
    def __contains__(self, item):
        return item in self.components
    
    def __iter__(self):
        return iter(self.components)
    
    def __getitem__(self, item):
        if isinstance(item, slice):
            path_length = len(self.components)
            slice_range = xrange(*item.indices(path_length))
            absolute = bool(self.absolute and
                            (len(slice_range) == 0 or slice_range[0] == 0))
            directory = bool(self.directory and
                             len(slice_range) > 0 and
                             slice_range[-1] == path_length - 1)
            return Path(self.components[item],
                        absolute=absolute,
                        directory=directory)
        elif isinstance(item, (int, long)) or hasattr(item, '__index__'):
            return self.components[item]
        else:
            type_name = type(item).__name__
            raise TypeError("Index or slice expected (got %s)" % type_name)
    
    # Comparison
    
    def __hash__(self):
        return hash(str(self))
    
    def __eq__(self, other):
        if isinstance(other, Path):
            return (self.components == other.components and
                    self.absolute == other.absolute and
                    self.directory == other.directory)
        elif isinstance(other, basestring):
            return str(self) == other
        elif isinstance(other, (list, tuple)):
            return self.components == tuple(other)
        else:
            return NotImplemented
    
    def __ne__(self, other):
        if isinstance(other, Path):
            return (self.components != other.components or
                    self.absolute != other.absolute or
                    self.directory != other.directory)
        elif isinstance(other, basestring):
            return str(self) != other
        elif isinstance(other, (list, tuple)):
            return self.components != tuple(other)
        else:
            return NotImplemented
    
    def __gt__(self, other):
        if isinstance(other, Path):
            if self.absolute and not other.absolute:
                return True
            result = cmp(self.components, other.components)
            if result == 1:
                return True
            elif result == 0:
                return bool(self.directory and not other.directory)
            else:
                return False
        elif isinstance(other, basestring):
            return bool(str(self) > other)
        elif isinstance(other, (list, tuple)):
            return self.components > tuple(other)
        else:
            return NotImplemented
    
    def __lt__(self, other):
        if isinstance(other, Path):
            if not self.absolute and other.absolute:
                return True
            result = cmp(self.components, other.components)
            if result == -1:
                return True
            elif result == 0:
                return bool(not self.directory and other.directory)
            else:
                return False
        elif isinstance(other, basestring):
            return bool(str(self) < other)
        elif isinstance(other, (list, tuple)):
            return self.components < tuple(other)
        else:
            return NotImplemented

### ASYNCHRONOUS PROGRAMMING ###

class ProgressDeferred(object):
    """
    A Deferred that can be called multiple times.
    
    This provides a similar API as Deferred, but allows callbacks to be called
    multiple times.  This is useful for an ongoing process.
    """
    def __init__(self):
        self._callbacks = []
    
    def callback(self, result):
        """
        Calls all the current callbacks with a value.
        
        :Parameters:
            result
                The first parameter to the first callback
        """
        for callback, callbackArgs, callbackKeywords in self._callbacks:
            result = callback(result, *callbackArgs, **callbackKeywords)
    
    def addCallback(self, callback, *callbackArgs, **callbackKeywords):
        """
        Adds a callback to call on progress.
        
        Any additional parameters are passed directly to the callback.
        
        :Parameters:
            callback
                A function to call
        """
        self._callbacks.append((callback, callbackArgs, callbackKeywords))

### LOGGING ###

LOG_USAGE = 25

class BelowLevelFilter(logging.Filter):
    def __init__(self, level, inclusive=True):
        self.level = level
        self.inclusive = inclusive
    
    def filter(self, record):
        if self.level == logging.NOTSET:
            return True
        elif self.inclusive:
            return record.levelno <= self.level
        else:
            return record.levelno < self.level

def setup_log(verbosity=logging.INFO, log_file=None):
    def _make_handler(format=None, level=None, below_level=None,
                      below_inclusive=True, force_stream=False):
        if force_stream or log_file is None:
            handler = logging.StreamHandler()
        else:
            handler = logging.FileHandler(log_file)
        if level is not None:
            handler.setLevel(level)
        if below_level is not None:
            handler.addFilter(BelowLevelFilter(below_level, below_inclusive))
        if format is not None:
            handler.setFormatter(logging.Formatter(format))
        return handler
    # Set up log parameters
    root_log = logging.getLogger('zfilenet')
    root_log.setLevel(verbosity)
    logging.addLevelName(LOG_USAGE, "USAGE")
    # Add main handler
    default_handler = _make_handler(None, logging.INFO, logging.WARNING, False)
    root_log.addHandler(default_handler)
    # Add "DEBUG:" to beginning of all debug messages
    debug_handler = _make_handler("DEBUG [%(name)s]: %(message)s",
                                  None, logging.DEBUG)
    root_log.addHandler(debug_handler)
    # Add "Warning:" to beginning of all warning messages
    warning_handler = _make_handler("Warning: %(message)s",
                                    logging.WARNING, logging.ERROR, False)
    root_log.addHandler(warning_handler)
    # Add "ERROR:" to beginning of all error messages
    error_format = "ERROR: %(message)s"
    error_handler = _make_handler(error_format, logging.ERROR)
    root_log.addHandler(error_handler)
    # Make sure that errors are sent to stderr
    if log_file is not None:
        stream_error_handler = _make_handler(error_format, logging.ERROR,
                                             force_stream=True)
        root_log.addHandler(stream_error_handler)

def _verbose_option(option, opt, value, parser):
    parser.values.verbosity -= 10
    if parser.values.verbosity <= 0:
        parser.values.verbosity = 1

def _quiet_option(option, opt, value, parser):
    parser.values.verbosity += 10

def add_log_options(parser):
    parser.set_defaults(verbosity=logging.INFO,
                        log_file=None,)
    group = parser.add_option_group("Logging")
    group.add_option('-v', '--verbose',
                     action='callback',
                     callback=_verbose_option,
                     nargs=0,
                     help="Increases the verbosity of the program.")
    group.add_option('-q', '--quiet',
                     action='callback',
                     callback=_quiet_option,
                     nargs=0,
                     help="Decreases the verbosity of the program.")
    group.add_option('--log-file',
                     action='store',
                     dest='log_file',
                     metavar="FILE",
                     help="Writes most logging information to a file, instead "
                          "of stderr. (errors are still written to stderr)")

def setup_log_options(options):
    setup_log(options.verbosity,
              log_file=options.log_file,)

### MISCELLANEOUS UTILITIES ###

def to_uuid(value):
    """
    Converts an object to a UUID.
    
    :Parameters:
        value
            A value to convert to a UUID
    :Raises TypeError: If the value could not be converted
    :Returns: The UUID equivalent of the value
    :ReturnType: uuid.UUID
    """
    if isinstance(value, UUID) or value is None:
        return value
    elif isinstance(value, basestring):
        if len(value) == 16:
            return UUID(bytes=value)
        else:
            return UUID(value)
    elif isinstance(value, (int, long)):
        return UUID(int=value)
    elif isinstance(value, (list, tuple)):
        return UUID(fields=value)
    else:
        raise TypeError("Unrecognized type for UUID, got '%s'" %
                        (type(value).__name__))

def data2hex(data_string):
    """
    Converts a byte string to a hex string.
    
    :Parameters:
        data_string : str
            The raw data
    :Returns: The encoded string
    :ReturnType: str
    """
    return ''.join('%.2x' % ord(b) for b in data_string)

def hex2data(hex_string):
    """
    Converts a hex string to a byte string.
    
    :Parameters:
        hex_string : str
            The encoded data
    :Returns: The decoded string
    :ReturnType: str
    """
    result = ''
    for i in xrange(len(hex_string) // 2):
        hex_unit = hex_string[i * 2] + hex_string[i * 2 + 1]
        result += chr(int(hex_unit, 16))
    return result

_qname_pattern = re.compile(r'^(?:\{([^}]*)\})?(.*)$')
def parse_qname(name):
    match = _qname_pattern.match(name)
    return match.groups()

def format_file_size(size):
    """
    Formats a file size to something human-readable.
    
    :Parameters:
        size : int
            The file size (in bytes)
    :Returns: The human-readable file size
    :ReturnType: unicode
    """
    base = 1024
    if size < (base ** 1):
        return u"%i bytes" % (size)
    elif size < (base ** 2):
        return u"%.1f KiB" % (size / (base ** 1))
    elif size < (base ** 3):
        return u"%.1f MiB" % (size / (base ** 2))
    elif size < (base ** 4):
        return u"%.1f GiB" % (size / (base ** 3))
    elif size < (base ** 5):
        # Not sure why we need this, but you never know...
        return u"%.1f TiB" % (size / (base ** 4))
    else:
        # Okay, we really shouldn't be worrying about sizes this big...
        return u"%.1f PiB" % (size / (base ** 5))

def format_datetime(dt):
    """
    Formats a date and time to ISO format.
    
    :Parameters:
        dt : datetime.datetime
            The date and time
    :Returns: The ISO format date
    :ReturnType: str
    """
    dt = dt.replace(microsecond=0)
    if dt.tzinfo is None:
        return dt.isoformat() + 'Z'
    else:
        return dt.isoformat()
